"""
Class Nadap
"""

from nadap.base import PreProcessingFlag, NOFLAG, ValEnv
from nadap.errors import DataValidationError
import nadap.references
import nadap.schema
import nadap.results


class Nadap:
    """
    Class Nadap
    """

    def __init__(self):
        self._references = nadap.references.References()
        self._validation_findings = nadap.results.FindingList()
        self._schema = None

    @property
    def schema(self) -> "nadap.schema.Schema":
        """
        Current loaded schema.
        """
        if self._schema is None:
            raise RuntimeError("Schema hasn't been set")
        return self._schema

    @schema.setter
    def schema(self, schema: nadap.schema.Schema | dict):
        if isinstance(schema, nadap.schema.Schema):
            self._schema = schema
        elif isinstance(schema, dict):
            self._schema = nadap.schema.Schema()
            self._schema.load_definition(schema)
        else:
            raise ValueError("Schema must be dict or Schema")
        self._root = self._schema.root

    @property
    def findings(
        self,
    ) -> "list[nadap.references.ReferenceFinding|nadap.results.ValidationFinding]":
        """
        Get all findings by previous validations
        """
        return (
            self._validation_findings + self._references.get_producer_consumer_issues()
        )

    def load_schema_definition(self, definition: dict) -> "nadap.schema.Schema":
        """
        Load schema definition and returns the initialized schema object.
        """
        self.schema = definition
        return self._schema

    def switch_namespace(self, namespace: str):
        """
        Switch to given namespace.
        Create namespace if not used before
        """
        self._references.change_namespace(namespace)
        self._validation_findings.namespace = namespace

    def validate(self, data: any, flags: PreProcessingFlag = NOFLAG) -> any:
        """
        Validate and pre-process data
        Given data remains unchanged.
        Returned data is pre-processed
        """
        if not isinstance(flags, PreProcessingFlag):
            raise ValueError("flags must be nadap pre-processing flags")
        try:
            r_data = self.schema.root.validate(
                data=data,
                path="",
                env=ValEnv(self._references, self._validation_findings, flags=flags),
            )
        except DataValidationError as e:
            raise DataValidationError(
                "Data Validation failed. Get findings for details."
            ) from e
        return r_data



"""
Results and finding classes
"""

# pylint: disable=too-few-public-methods


class Finding:
    """
    Base class for all findings
    """

    def __init__(self, message: str, path: str = None) -> None:
        self.path = path if path else ""
        self.message = message
        self.namespace = ""

    def __str__(self) -> str:
        _str = self.message
        if self.path:
            _str = self.path + ": " + _str
        if self.namespace:
            _str = self.namespace + " > " + _str
        return _str

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        return str(self) > str(other)

    def __ge__(self, other):
        return self > other or self == other

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return self < other or self == other


class ValidationFinding(Finding):
    """
    Class to handle a finding during data validation process
    """


class FindingList(list):
    """
    List containing all findings
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.namespace = None

    def append(self, f: "Finding"):
        """
        Append new finding and change namespace within finding.
        """
        if self.namespace:
            f.namespace = self.namespace
        super().append(f)

    def extend(self, f: "list[Finding]"):
        """
        Append new finding and change namespace within finding.
        """
        if self.namespace:
            for i in f:
                i.namespace = self.namespace
        super().extend(f)



"""
Integer data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.mixin.not_allowed_value
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value
from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError

DOC_DT_NAME = "Integer/String"
DOC_DT_DESCRIPTION = """
An **str_int** data type tests data for being an instance of
python's built-in class `int` or class `str` and string must
represent an integer value.

> Data is converted to `int` by default.
> Use `convert_to` option to enforce conversion to other data types.
"""
DOC_DT_FEATURES = f"""
- Validate minimal and maximal integer value{nadap.mixin.ranges.DOC_FEATURES}
- Validate against allowed and not allowed integer values
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: str_int
description: "Example str_int definition"

minimum: 1
maximum: 4096

not_allowed_values:
  - 1001
  - "1002"

reference:
  key: ref_key
  mode: consumer
"""


class StrInt(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    Integer/String datatype class
    """

    data_type_name = "str_int"
    _cls_python_classes = [int, str]
    _convert_to_classes = {"float": float, "str": str}
    _doc_data_type = "int|str[int]"

    def _test_data_type(self, data: any, path: str, env: "ValEnv" = None):
        super()._test_data_type(data=data, path=path, env=env)
        msg = ""
        try:
            data = int(data)
        except ValueError:
            msg = "String does not represent an integer value"
        if msg:
            if env is None:
                raise SchemaDefinitionError(
                    msg,
                    path,
                )
            self._create_finding_with_error(msg, path, env)
        return data


DOC_DT_CLASS = StrInt  # pylint: disable=invalid-name



"""
Integer data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.mixin.accuracy
import nadap.mixin.not_allowed_value
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value
from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError

DOC_DT_NAME = "Number/String"
DOC_DT_DESCRIPTION = """
An **str_number** data type tests data for being an instance of
python's built-in class `int`, `float` or class `str` and string must
represent an integer or a float value.

> String is converted to `float` if representing an float value
> or to `int` if representing a integer value
> Use `convert_to` option to enforce convertion to other data types.
"""
DOC_DT_FEATURES = f"""
- Validate minimal and maximal number value{nadap.mixin.ranges.DOC_FEATURES}
- Validate against allowed and not allowed number values
- Validate if float value meets allowed accuracy
- Conversion supports rounding the float value
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: str_number
description: "Example str_number definition"

minimum: 1
maximum: 3.0

not_allowed_values:
  - 3.0
  - 1
  - "0.9"
  - "4"

reference: ref_key
"""


class StrNumber(
    nadap.mixin.accuracy.AccuracyMixin,
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    Number/String datatype class
    """

    data_type_name = "str_number"
    _cls_python_classes = [int, float, str]
    _convert_to_classes = {"float": float, "int": int, "str": str}
    _doc_data_type = "float|int|str[float|int]"

    def _test_data_type(self, data: any, path: str, env: "ValEnv" = None):
        super()._test_data_type(data=data, path=path, env=env)
        if isinstance(data, float):
            return data
        msg = ""
        try:
            r_data = float(data)
            try:
                r_data = int(data)
            except ValueError:
                pass
        except ValueError:
            msg = "String does not represent an integer or a float value"
        if msg:
            if env is None:
                raise SchemaDefinitionError(msg, path)
            self._create_finding_with_error(msg, path, env)
        return r_data


DOC_DT_CLASS = StrNumber  # pylint: disable=invalid-name



"""
Number data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base
import nadap.mixin.accuracy
import nadap.mixin.not_allowed_value
import nadap.mixin.min_max
import nadap.mixin.ranges
import nadap.mixin.allowed_value

DOC_DT_NAME = "Number"
DOC_DT_DESCRIPTION = """
A **number** data type tests data for being an instance of
python's built-in class `float` or `int`.

> If using *number* data type with allowed values defined, check *enum* data type as a better option.
"""
DOC_DT_FEATURES = f"""
- Validate minimal and maximal data value{nadap.mixin.ranges.DOC_FEATURES}
- Validate against allowed and not allowed number values
- Validate if float value meets allowed accuracy
- Conversion supports rounding the float value
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: number
description: "Example number definition"

minimum: 1
maximum: 3.1

not_allowed_values:
  - 3.0
  - 1
  - 0.9

reference: ref_key
"""


class Number(
    nadap.mixin.accuracy.AccuracyMixin,
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.mixin.min_max.MinMaxValueMixin,
    nadap.types.base.BaseType,
):
    """
    Number datatype class
    Includes integer and floats
    """

    data_type_name = "number"
    _cls_python_classes = [int, float]
    _convert_to_classes = {"str": str, "float": float, "int": int}
    _doc_data_type = "float|int"


DOC_DT_CLASS = Number  # pylint: disable=invalid-name



"""
IPv4 Network Data Type Class
"""

# pylint: disable=too-few-public-methods

import ipaddress

import nadap.mixin.ip
import nadap.types.base

DOC_DT_NAME = "IPv4 Network"
DOC_DT_DESCRIPTION = """
An **ip4_network** data type tests data for being an instance of
python's built-in class `str` and the string represents an IPv4
network.
"""
DOC_DT_FEATURES = """
- Validate IP specific attributes
- Validate against allowed and not allowed IP networks
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: ip4_network
description: "Example IPv4 network definition"

multicast: False
private: True

not_allowed_values:
  - 10.20.0.0/16
  - 192.168.0.0/24

reference:
  key: ref_key
"""


class Ip4Network(
    nadap.mixin.ip.Ip4NetworkMixin, nadap.mixin.ip.IpMixin, nadap.types.base.BaseType
):
    """
    Add handling of strings representing IPv4 networks
    """

    _ip_data_type_name = "IPv4 network"

    def _get_ip_obj(self, data):
        try:
            ipaddress.IPv4Address(data)
        except ValueError:
            return ipaddress.IPv4Network(data)
        raise ValueError()


DOC_DT_CLASS = Ip4Network  # pylint: disable=invalid-name



"""
ID list class
"""

# pylint: disable=too-few-public-methods

import nadap.types.list
from nadap.base import ValEnv
from nadap.doc import (
    TableRow,
    Type,
    TextList,
)
from nadap.errors import DataValidationError

DOC_DT_NAME = "IDList"
DOC_DT_DESCRIPTION = """
An **idlist** data type represents a python dictionary with IDs as keys.
It tests data for being an instance of python's built-in class `dict`.
"""
DOC_DT_FEATURES = """
- Validate number of IDs/keys
- Validate IDs/keys with nested data type definition
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: idlist
description: |
  Example idlist definition for string IDs with integer values and
  enhanced referencing.
default_value:
  nadap: 1
  rulez!: 2

minimum: 2
maximum: 15

id:
  type: str
  reference: ref_key_id
elements:
  type: int
  reference: ref_key_element
"""


class IDList(
    nadap.types.list.ListBase,
):
    """
    ID list datatype class (dictionary with IDs as keys)
    """

    data_type_name = "idlist"
    _cls_python_classes = [dict]
    _doc_data_type = "dict"

    def __init__(self, **kwargs):
        self.id_data_type = None
        super().__init__(**kwargs)

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        self.id_data_type = self._schema.load_data_type_by_definition(
            self.id_data_type, f"{schema_path}.id"
        )

    def _pop_options(self, definition: dict, schema_path: str):
        self.id_data_type = definition.pop("id", "str")
        super()._pop_options(definition, schema_path)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        ret_dict = {}
        for index, k in enumerate(data):
            try:
                r_key = self.id_data_type.validate(
                    data=k,
                    path=f"{path}{{#{index}}}",
                    env=env,
                )
                r_data = self.elements_data_type.validate(
                    data=data[k],
                    path=f"{path}.{k}",
                    env=env,
                )
            except DataValidationError:
                ret_dict = None
            else:
                if ret_dict is not None:
                    ret_dict[r_key] = r_data
        if ret_dict is None:
            raise DataValidationError()
        return ret_dict

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get doc table entries for this data type definition
        """

        def _structure_path_element():
            if dict in self.elements_data_type.python_classes:
                return "."
            return ""

        ret_list = super().doc_table_rows
        element_type = None

        id_entries = self.id_data_type.doc_table_rows
        id_entries[0].type = Type("")
        id_entries[0].variable = self.id_data_type.doc_value_name
        id_entries[0].variable.placeholder = True
        id_entries[0].key_value = True
        for e in id_entries:
            e.path = f"<{id_entries[0].variable.name}>" + e.path

        id_e_entries = self.elements_data_type.doc_table_rows
        if (
            len(id_e_entries) > 1
            and set(self.elements_data_type.python_classes).intersection({dict, list})
            and not id_e_entries[0].default
            and not id_e_entries[0].description
        ):
            element_type = Type(self.elements_data_type.doc_types)
            id_e_entries = id_e_entries[1:]
        for entry in id_e_entries:
            entry.indent.level += 1
            entry.path = (
                f"<{id_entries[0].variable.name}>{_structure_path_element()}"
                + entry.path
            )

        if not set(self.elements_data_type.python_classes).intersection({dict, list}):
            # id row and value info needs to be merged.
            id_entries[0].type = id_e_entries[0].type
            id_entries[0].default = id_e_entries[0].default
            if id_entries[0].restrictions:
                id_entries[0].restrictions = TextList(
                    ["Variable:", id_entries[0].restrictions.t]
                )
            else:
                id_entries[0].restrictions = TextList([])
            if id_e_entries[0].restrictions:
                id_entries[0].restrictions = id_entries[0].restrictions.t + TextList(
                    ["Value:", id_e_entries[0].restrictions.t]
                )
            if id_entries[0].description:
                id_entries[0].description = TextList(
                    ["Variable:", id_entries[0].description.t]
                )
            else:
                id_entries[0].description = TextList([])
            if id_e_entries[0].description:
                id_entries[0].description = id_entries[0].description.t + TextList(
                    ["Value:", id_e_entries[0].description.t]
                )
            ret_list += id_entries
        else:
            if element_type:
                id_entries[0].type = element_type
            ret_list += id_entries + id_e_entries
        return ret_list

    @property
    def doc_yaml(self) -> "list[str]":
        """
        Get data structure
        """
        if (
            dict in self.elements_data_type.python_classes
            or list in self.elements_data_type.python_classes
        ):
            return [f"<{self.id_data_type.yaml_data_type}>:"] + [
                f"  {x}" for x in self.elements_data_type.doc_yaml
            ]
        return [
            f"<{self.id_data_type.yaml_data_type}>: {self.elements_data_type.doc_yaml[0]}"
        ]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **id** | <code>dict&#124;str</code> | 'str' | | "
            + "ID's data type definition |",
            "| &nbsp;&nbsp;*option 1* | <code>str</code> |  | "
            + "ID's data type string (short definition) |",
            "| &nbsp;&nbsp;*option 2* | <code>dict</code> | | ID's data type definition |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "id: <dict|str>",
            "  # Mulitype!!!",
            "  <str>",
            "  # or",
            "  <dict>",
        ]


DOC_DT_CLASS = IDList  # pylint: disable=invalid-name



"""
BGP community data type class
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.types.base
import nadap.mixin.ranges
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value
import nadap.schema
from nadap.base import ValEnv, CONVERT_DATA
from nadap.errors import SchemaDefinitionError
from nadap.doc import UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField


DOC_DT_NAME = "BGP Community"
DOC_DT_DESCRIPTION = """
A **bgp_community** data type tests data for being an instance of
python's built-in class `str` or `int` and if it represents a BGP standard community (4 bytes).
"""
DOC_DT_FEATURES = """
- Validate BGP Community format
- Supports BGP Community format conversion
- Validate BGP Community specific attributes
- Validate against allowed and not allowed BGP Communities
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = r"""
type: bgp_community
description: "Example bgp_community definition"
default_value: GRACEFUL_SHUTDOWN

format: colon-separated
well_known: false
public: false
convert_to_format: integer

not_allowed_values:
  - NO_EXPORT
  - 7001
  - 0:7002
  - 0xA0001

not_allowed_ranges:
  - start: 65498:0
    end: 65498:65535

reference: ref_key
"""


NEW_FORMAT = "new-format"  # "65000:100"
INT_FORMAT = "integer"  # 11223344
HEX_FORMAT = "hex"  # FFFF0001 or ffff0001
WELL_KNOWN_MAP = {
    # Source:
    # https://www.iana.org/assignments/bgp-well-known-communities/bgp-well-known-communities.xhtml
    "GRACEFUL_SHUTDOWN": 4294901760,
    "ACCEPT_OWN": 4294901761,
    "ROUTE_FILTER_TRANSLATED_v4": 4294901762,
    "ROUTE_FILTER_v4": 4294901763,
    "ROUTE_FILTER_TRANSLATED_v6": 4294901764,
    "ROUTE_FILTER_v6": 4294901765,
    "LLGR_STALE": 4294901766,
    "NO_LLGR": 4294901767,
    "accept-own-nexthop": 4294901768,
    "BLACKHOLE": 4294902426,
    "NO_EXPORT": 4294967041,
    "NO_ADVERTISE": 4294967042,
    "NO_EXPORT_SUBCONFED": 4294967043,
    "NOPEER": 4294967044,
}


class BgpCommunity(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.types.base.BaseType,
):
    """
    BGP Community datatype class
    """

    data_type_name = "bgp_community"
    _cls_python_classes = [int, str]
    _convert_to_classes = {}
    _doc_data_type = "BGP Community"

    def __init__(self, **kwargs):
        self._format = None
        self._public = None
        self._well_known = None
        self._detected_format = None
        self._was_well_known_string = None
        self._convert_to_format = None
        super().__init__(**kwargs)

    @staticmethod
    def _check_format_str(format_, path):
        if format_ not in [NEW_FORMAT, INT_FORMAT, HEX_FORMAT]:
            raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                msg="Format unknown",
                path=path,
            )

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)

        if self._format is not None:
            nadap.schema.is_str(self._format, f"{schema_path}.format")
            self._check_format_str(self._format, f"{schema_path}.format")

        if self._convert_to_format is not None:
            self._check_format_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )

    def _pop_options(self, definition: dict, schema_path: str):
        self._format = definition.pop("format", self._format)
        self._public = definition.pop("public", self._public)
        self._well_known = definition.pop("well_known", self._well_known)
        self._convert_to_format = definition.pop(
            "convert_to_format", self._convert_to_format
        )
        super()._pop_options(definition, schema_path)

    def _validate_source_as(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        source_as = int(data / 2**16)
        msg = None
        if self._public is not None:
            if self._public:
                if 64512 <= source_as <= 65534:
                    msg = "BGP community's source AS number is within private space"
            elif 1 <= source_as < 64512:
                msg = "BGP community's source AS number is within public space"
        self._raise_exceptions(messages=msg, path=path, env=env)

    def _validate_well_known(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ):
        msg = None
        if self._well_known is None:
            return
        if self._well_known:
            if data not in WELL_KNOWN_MAP.values():
                msg = "BGP community is not a well-known community"
        elif data in WELL_KNOWN_MAP.values():
            msg = "BGP community is a well-known community"
        self._raise_exceptions(messages=msg, path=path, env=env)

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        if fmt == NEW_FORMAT:
            source_as = int(data / 2**16)
            operator_value = data % 2**16
            return f"{source_as}:{operator_value}"
        if fmt == HEX_FORMAT:
            return hex(data).upper().replace("0X", "0x")
        if self._was_well_known_string:
            return self._was_well_known_string
        return data

    def _detect_data_format(self, data) -> str:
        if isinstance(data, int):
            return INT_FORMAT
        if isinstance(data, str):
            if data.count(":") == 1:
                return NEW_FORMAT
            if data.startswith("0x"):
                return HEX_FORMAT
        return None

    def _validate_format(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> str:
        # pylint: disable=too-many-branches
        if isinstance(data, int):
            # BGP AS number is in asplain format.
            if self._format and self._format != INT_FORMAT:
                self._raise_exceptions(
                    f"BGP community is not in {self._format} format", path, env
                )
            self._detected_format = INT_FORMAT
            return data
        if self._format == INT_FORMAT:
            self._raise_exceptions(
                f"BGP community is not in {INT_FORMAT} format", path, env
            )

        if data.count(":") == 1:
            # BGP community is in new format.
            self._detected_format = NEW_FORMAT
            source_as, operator_value = data.split(":")
            try:
                source_as = int(source_as)
                operator_value = int(operator_value)
            except ValueError:
                self._raise_exceptions("Invalid BGP community format", path, env)
            if self._format and self._format == HEX_FORMAT:
                self._raise_exceptions(
                    f"BGP community is not in {self._format} format", path, env
                )
            return source_as * 2**16 + operator_value

        # BGP community is not in new format
        if len(data) == 0:
            self._raise_exceptions(
                "Empty string is not allowed as a BGP community", path, env
            )
        if data in WELL_KNOWN_MAP:
            self._was_well_known_string = data
            return WELL_KNOWN_MAP[data]
        self._detected_format = HEX_FORMAT
        if data.startswith("0x"):
            try:
                int_data = int(data, 16)
                return int_data
            except ValueError:
                pass
        self._raise_exceptions("String is no valid BGP community", path, env)
        return None

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if CONVERT_DATA in env.flags and self._convert_to_format is not None:
            data = self._apply_data_format(data, self._convert_to_format)
        else:
            data = self._apply_data_format(data, self._detected_format)
        return data

    def _test_data_type(self, data: str, path: str, env: "ValEnv" = None) -> "str":
        data = super()._test_data_type(data=data, path=path, env=env)
        data = self._validate_format(data=data, path=path, env=env)
        if not 0 <= data <= 4294967295:
            self._raise_exceptions(
                "BGP community value is out of allowed range 0-4294967295", path, env
            )
        self._validate_well_known(data=data, path=path, env=env)
        self._validate_source_as(data=data, path=path, env=env)
        return data

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        _restrictions = UnorderedTextList()
        if self._format:
            _restrictions.append((f"must be in format '{self._format}'"))
        if self._public is not None:
            _restrictions.append(
                f"Source AS must be {'public' if self._public else 'private'}"
            )
        if self._well_known is not None:
            _restrictions.append(
                f"must{'' if self._well_known else ' not'} be a well-known BGP community"
            )
        if _restrictions:
            tf.append("BGP Community:")
            tf.append(_restrictions)
        return tf

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._doc_data_type

    @property
    def doc_types(self) -> list[str]:
        """Get list of data type strings"""
        return [self._doc_data_type]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        allowed_formats = (
            "allowed_values:<br>"
            + f"{cls._markdown_indent}- {INT_FORMAT}<br>"
            + f"{cls._markdown_indent}- {NEW_FORMAT}<br>"
            + f"{cls._markdown_indent}- {HEX_FORMAT}"
        )
        f_desc = (
            f"- {INT_FORMAT} example: 70116<br>"
            + f"- {NEW_FORMAT} example: 1.4580<br>"
            + f"- {HEX_FORMAT} example: '0x111E4'"
        )
        return super()._doc_options_md_upper_part() + [
            f"| **format** | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            "| **public** | <code>bool</code> | | | "
            + "BGP Community's source AS must be a public (true) or a private (false) AS number |",
            "| **well_known** | <code>bool</code> | | | "
            + "BGP Community must (true) or must not (false) be well-known |",
            f"| **convert_to_format** | <code>enum</code> | | {allowed_formats} | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"format: <{INT_FORMAT}|{NEW_FORMAT}|{HEX_FORMAT}>",
            "public: <true|false>",
            "well_known: <true|false>",
            f"convert_to_format: <{INT_FORMAT}|{NEW_FORMAT}|{HEX_FORMAT}>",
        ]


DOC_DT_CLASS = BgpCommunity  # pylint: disable=invalid-name



"""
IP Interface Data Type Class
"""

# pylint: disable=too-few-public-methods

import ipaddress

import nadap.mixin.ip
import nadap.types.base

DOC_DT_NAME = "IP Interface"
DOC_DT_DESCRIPTION = """
An **ip_interface** data type tests data for being an instance of
python's built-in class `str` and the string represents an IPv4
or IPv6 interface.
"""
DOC_DT_FEATURES = """
- Validate IP specific attributes
- Validate against allowed and not allowed IP interfaces
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: ip_interface
description: "Example IP interface definition"

multicast: False
private: True

not_allowed_values:
  - 10.10.0.1/16
  - 192.168.0.1/24
  - 2001:cafe::1/64

reference:
  key: ref_key
"""


class IpInterface(nadap.mixin.ip.IpMixin, nadap.types.base.BaseType):
    """
    Add handling of strings representing IPv4 or IPv6 interfaces
    """

    _ip_data_type_name = "IP interface"

    def _get_ip_obj(self, data):
        try:
            ipaddress.ip_address(data)
        except ValueError:
            try:
                ip_a = ipaddress.ip_network(data)
                if isinstance(ip_a, ipaddress.IPv4Network) and ip_a.prefixlen in [
                    31,
                    32,
                ]:
                    return ipaddress.IPv4Interface(data)
                if isinstance(ip_a, ipaddress.IPv6Network) and ip_a.prefixlen in [
                    127,
                    128,
                ]:
                    return ipaddress.IPv6Interface(data)
            except ValueError:
                return ipaddress.ip_interface(data)
        raise ValueError()


DOC_DT_CLASS = IpInterface  # pylint: disable=invalid-name



"""
Dict class
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

from collections import OrderedDict
import copy
import nadap.results
import nadap.types.base
import nadap.mixin.regex_mode
import nadap.mixin.replace_empty_to
import nadap.schema
from nadap.doc import TableRow
from nadap.base import UNDEFINED, number_to_str_number, ValEnv
from nadap.errors import SchemaDefinitionError, DataValidationError

if TYPE_CHECKING:
    from nadap.doc import DocTableRow

DOC_DT_NAME = "Dictionary"
DOC_DT_DESCRIPTION = """
A **dict** data type tests data for being an instance of
python's built-in class `dict`.
"""
DOC_DT_FEATURES = """
- Check data type of key values.
- **Regex Mode** matches keys based on regex patterns.
- Advanced data key validation:
  - Check data for **required** keys
  - Validate **restrictions** among data keys like *requires*, *excludes* or *xor*
- Supports **Referencing Feature**. Adds an option to filter key/value pairs used as reference value
"""
DOC_DT_YAML_EXAMPLE = """
type: dict
description: "Example dict definition with regex mode enabled and some key restrictions in place"
default_value:
  a11: 1
reference:
references:
  - 'unique_key'
  - key: 'consumer_key'
    mode: consumer
    keys:
      - b11  # If key b11 is in data it is included in reference data
      - 12  # If key 12 is in data it is included in reference data

keys:
  11: str
  a11: int
  a12: int
  12: str
  b11: float
  b12: float

restrictions:
  keys:
    b11:
      requires:
        - a11  # If key b11 in data key a11 must be in data, too
      excludes:
        - b12  # If key b11 in data key b12 mustn't be in data
  required:
    - a11  # Key 'a11' must be in data
  xor_required:
    -  # If key 11 is in data key 12 mustn't be in data - and vice-versa
      - 11
      - 11
"""


class Dict(
    nadap.mixin.replace_empty_to.ReplaceEmptyToMixin,
    nadap.mixin.regex_mode.RegexModeMixin,
    nadap.types.base.BaseType,
):
    """
    Dict datatype class
    """

    data_type_name = "dict"
    _cls_python_classes = [dict]
    _support_replace_empty_to = True
    _doc_data_type = "dict"

    def __init__(self, **kwargs):
        self.keys = None
        self.regex_store = {}
        self.key_restrictions = {}
        self.required_keys = []
        self.xor_required = []
        self.ignore_unknown_keys = False
        self.reference_keys = []
        super().__init__(**kwargs)

    def _check_for_defined_key(self, a_name: any, schema_path: str):
        if not self.regex_mode and a_name not in self.keys:
            raise SchemaDefinitionError("Key not defined", schema_path)

    def _key_regex(self, a_name: str, schema_path: str):
        if self.regex_mode:
            self.regex_store[a_name] = nadap.schema.compile_regex_string(
                pattern=a_name,
                multiline=self.regex_multiline,
                fullmatch=self.regex_fullmatch,
                schema_path=schema_path,
            )

    def _validate_keys_option(self, schema_path: str):
        if not self.keys:
            return
        nadap.schema.is_dict(self.keys, schema_path)
        keys_definition = self.keys
        self.keys = OrderedDict()
        for a_name, a_definition in keys_definition.items():
            a_path = f"{schema_path}.{a_name}"
            self._key_regex(a_name, a_path)
            self.keys[a_name] = self._schema.load_data_type_by_definition(
                definition=a_definition, path=a_path
            )

    def _parse_key_restrictions(self, restrictions: dict, schema_path: str):
        nadap.schema.is_dict(restrictions, schema_path)
        restrictions = copy.deepcopy(restrictions)
        requires = restrictions.pop("requires", [])
        req_path = f"{schema_path}.requires"
        nadap.schema.is_list(requires, req_path)
        for index, a_name in enumerate(requires):
            _path = f"{req_path}[{index}]"
            self._check_for_defined_key(a_name, _path)
            self._key_regex(a_name, _path)
        excludes = restrictions.pop("excludes", [])
        ex_path = f"{schema_path}.excludes"
        nadap.schema.is_list(excludes, ex_path)
        for index, a_name in enumerate(excludes):
            _path = f"{ex_path}[{index}]"
            self._check_for_defined_key(a_name, _path)
            self._key_regex(a_name, _path)
        nadap.schema.no_more_definition_options(
            definition=restrictions,
            source="key restrictions",
            path=schema_path,
        )

    def _validate_restrictions_option(self, schema_path: str):
        if not self.key_restrictions:
            return
        if not self.keys:
            raise SchemaDefinitionError(
                "Not allowed if no keys are defined",
                f"{schema_path}.restrictions",
            )
        nadap.schema.is_dict(self.key_restrictions, schema_path)
        restrictions = self.key_restrictions
        # Parse required_keys:
        req_attr = restrictions.pop("required", [])
        if req_attr == "all":
            self.required_keys = list(self.keys)
        else:
            nadap.schema.is_list(req_attr, f"{schema_path}.required")
            for index, a_name in enumerate(req_attr):
                _path = f"{schema_path}.required[{index}]"
                self._check_for_defined_key(a_name, _path)
                self._key_regex(a_name, _path)
                self.required_keys.append(a_name)
        # Parse key restrictions
        self.key_restrictions = restrictions.pop("keys", {})
        a_path = f"{schema_path}.keys"
        nadap.schema.is_dict(self.key_restrictions, a_path)
        for a_name, a_restrictions in self.key_restrictions.items():
            a_n_path = f"{a_path}.{a_name}"
            self._check_for_defined_key(a_name, a_n_path)
            self._key_regex(a_name, a_n_path)
            self._parse_key_restrictions(
                restrictions=a_restrictions,
                schema_path=a_n_path,
            )
        # Parse xor_required:
        self.xor_required = restrictions.pop("xor_required", [])
        xor_path = f"{schema_path}.xor_required"
        nadap.schema.is_list(self.xor_required, xor_path)
        if self.xor_required and set(self.required_keys) == set(self.keys):
            raise SchemaDefinitionError(
                "Not allowed if all keys are required", xor_path
            )
        for index1, xor_list in enumerate(self.xor_required):
            xor_i_path = f"{xor_path}[{index1}]"
            nadap.schema.is_list(xor_list, xor_i_path)
            if len(xor_list) < 2:
                raise SchemaDefinitionError(
                    "List must have at least two key names", xor_i_path
                )
            for index2, a_name in enumerate(xor_list):
                _path = f"{xor_i_path}[{index2}]"
                self._check_for_defined_key(a_name, _path)
                self._key_regex(a_name, _path)
        nadap.schema.no_more_definition_options(
            definition=restrictions, source="restrictions", path=schema_path
        )

    def _convert_reference_data(self, data: any) -> any:
        if not self.reference_keys:
            return data
        ret_data = {}
        if self.regex_mode:
            for r_name in self.reference_keys:
                for a_name in data:
                    if self.regex_store[r_name].match(a_name):
                        ret_data[a_name] = data[a_name]
        else:
            for a_name in self.reference_keys:
                if a_name in data:
                    ret_data[a_name] = data[a_name]
        return ret_data

    def _pop_reference_definition(self, ref_definition: dict, schema_path: str):
        if "keys" in ref_definition:
            a_list = ref_definition.pop("keys")
            _path = f"{schema_path}.keys"
            nadap.schema.is_non_empty_list(a_list, _path)
            for index, a_name in enumerate(a_list):
                i_path = f"{_path}[{index}]"
                self._check_for_defined_key(a_name, i_path)
                self._key_regex(a_name, i_path)
            self.reference_keys = a_list

        super()._pop_reference_definition(
            ref_definition=ref_definition, schema_path=schema_path
        )

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        self._validate_keys_option(
            schema_path=f"{schema_path}.keys",
        )
        self._validate_restrictions_option(schema_path=f"{schema_path}.restrictions")

    def _pop_options(self, definition: dict, schema_path: str):
        self.ignore_unknown_keys = definition.pop(
            "ignore_unknown_keys", "keys" not in definition
        )
        self.keys = definition.pop("keys", {})
        self.key_restrictions = definition.pop("restrictions", {})
        super()._pop_options(definition, schema_path)

    def _get_matching_regex_string(
        self, a_name: any, regex_strings: list[str]
    ) -> str | None:
        for re_str in regex_strings:
            re_obj = self.regex_store[re_str]
            if re_obj.match(a_name):
                return re_str
        return None

    def _test_a_name_in_data(self, a_name, data: dict) -> bool:
        if self.regex_mode:
            for data_key in data:
                if self.regex_store[a_name].match(data_key):
                    return True
        elif a_name in data:
            return True
        return False

    def _test_key_data(self, data, path, env):
        # pylint: disable=too-many-branches
        r_dict = {}
        for a_name in data:
            if path:
                if path[-1] != "]":
                    key_path = f"{path}.{a_name}"
                else:
                    key_path = path + a_name
            else:
                key_path = a_name
            dt = None
            if self.regex_mode:
                def_key = self._get_matching_regex_string(a_name, list(self.keys))
                if def_key is not None:
                    dt = self.keys[def_key]
            elif a_name in self.keys:
                dt = self.keys[a_name]
            if dt:
                try:
                    a_data = dt.validate(
                        data=data[a_name],
                        path=key_path,
                        env=env,
                    )
                except DataValidationError:
                    r_dict = None
                else:
                    if r_dict is not None:
                        r_dict[a_name] = a_data
            else:
                if self.ignore_unknown_keys:
                    if r_dict is not None:
                        r_dict[a_name] = data[a_name]
                else:
                    env.findings.append(
                        nadap.results.ValidationFinding("Key not allowed", key_path)
                    )
                    r_dict = None
        if r_dict is None:
            raise DataValidationError()
        return r_dict

    def _test_required_keys(self, data, path, env):
        found_required = set()
        for a_name in self.required_keys:
            if self.regex_mode:
                for data_key in data:
                    if self.regex_store[a_name].match(data_key):
                        found_required.add(a_name)
                        break
            elif a_name in data:
                found_required.add(a_name)
        if missing_a_names := set(self.required_keys) - found_required:
            for a_name in missing_a_names:
                if self.regex_mode:
                    msg = f"No key found matching regex pattern '{a_name}'"
                else:
                    msg = f"Required key '{a_name}' not found"
                env.findings.append(nadap.results.ValidationFinding(msg, path))
            raise DataValidationError()

    def _test_required_xor(self, data, path, env):
        for a_list in self.xor_required:
            first_found = None
            for a_name in a_list:
                if self._test_a_name_in_data(a_name, data):
                    if first_found:
                        self._create_finding_with_error(
                            msg=f"Key {first_found} excludes key {a_name}",
                            path=path,
                            env=env,
                        )
                    first_found = "a_name"
            if not first_found:
                self._create_finding_with_error(
                    msg=f"One of these keys required: {', '.join(a_list)}",
                    path=path,
                    env=env,
                )

    def _test_key_restrictions(self, data, path, env):
        invalid = False
        for a_name in data:
            _path = f"{path}.{a_name}"
            a_restrictions = {}
            if self.regex_mode:
                if re_str := self._get_matching_regex_string(
                    a_name=a_name, regex_strings=list(self.key_restrictions)
                ):
                    a_restrictions = self.key_restrictions[re_str]
            else:
                a_restrictions = self.key_restrictions.get(a_name, {})
            if a_restrictions:
                for aa_name in a_restrictions.get("requires", []):
                    if not self._test_a_name_in_data(aa_name, data):
                        env.findings.append(
                            nadap.results.ValidationFinding(
                                f"Requires key '{aa_name}'", _path
                            )
                        )
                        invalid = True
                for aa_name in a_restrictions.get("excludes", []):
                    if self._test_a_name_in_data(aa_name, data):
                        env.findings.append(
                            nadap.results.ValidationFinding(
                                f"Excludes key '{aa_name}'", _path
                            )
                        )
                        invalid = True
        if invalid:
            raise DataValidationError()

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        invalid = False
        # Check required keys:
        try:
            self._test_required_keys(data=data, path=path, env=env)
        except DataValidationError:
            invalid = True
        # Check required_xor keys:
        try:
            self._test_required_xor(data=data, path=path, env=env)
        except DataValidationError:
            invalid = True
        # Check key-specific restrictions
        try:
            self._test_key_restrictions(data=data, path=path, env=env)
        except DataValidationError:
            invalid = True
        # Validate keys in data
        try:
            attr_data = self._test_key_data(
                data=data,
                path=path,
                env=env,
            )
        except DataValidationError:
            invalid = True
        if invalid:
            raise DataValidationError()
        return attr_data

    def _set_default(self, data: any):
        """
        Set values of nested data types to default if missing.
        """
        if self.regex_mode or not self.keys:
            return data
        for a_name, dt in self.keys.items():
            if a_name not in data and dt.default_value is not UNDEFINED:
                data[a_name] = dt.default_value
        return data

    def _doc_key_restrictions(self, a_name: any) -> "str|list":
        if self.required_keys and a_name in self.required_keys:
            a_restrictions = ["required"]
        else:
            a_restrictions = []
        requires = set()
        excludes = set()
        for xor_list in self.xor_required:
            if a_name in xor_list:
                _list = xor_list[:]
                _list.remove(a_name)
                excludes.update(_list)
        if a_name in self.key_restrictions:
            requires.update(self.key_restrictions[a_name].get("requires", []))
            excludes.update(self.key_restrictions[a_name].get("excludes", []))
        if requires or excludes:
            if requires:
                if len(requires) == 1:
                    a_restrictions.append(f"requires: {requires.pop()}")
                else:
                    a_restrictions.append("requires:")
                    a_restrictions += [
                        f" - {str(number_to_str_number(x))}" for x in sorted(requires)
                    ]
            if excludes:
                if len(excludes) == 1:
                    a_restrictions.append(f"excludes: {excludes.pop()}")
                else:
                    a_restrictions.append("excludes:")
                    a_restrictions += [
                        f" - {str(number_to_str_number(x))}" for x in sorted(excludes)
                    ]
        return a_restrictions

    def _format_key_table_entries(
        self,
        entries: "list[TableRow]",
        key: str,
        add_top_row: "bool" = False,
    ) -> "list[TableRow]":
        def _structure_path_element(path):
            if not path or path.startswith("["):
                return ""
            return "."

        if entries:
            entries[0].variable.name = key
            entries[0].restrictions.t = (
                self._doc_key_restrictions(key) + entries[0].restrictions.t
            )
            entries[0].path = (
                f"{key}{_structure_path_element(entries[0].path)}" + entries[0].path
            )
            for entry in entries[1:]:
                entry.indent.level += 1
                entry.path = f"{key}{_structure_path_element(entry.path)}" + entry.path
        if add_top_row:
            entries.insert(0, super().doc_table_rows[0])
        return entries

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get doc table entries for this data type definition
        """
        ret_list = super().doc_table_rows
        if self.keys:
            for a_name, obj in self.keys.items():
                ret_list += self._format_key_table_entries(
                    entries=obj.doc_table_rows, key=a_name
                )
        else:
            ret_list[0].empty_structure = True
        return ret_list

    def get_doc_table_rows_dict(
        self, max_level: int = 0, _level: int = 0
    ) -> "tuple[dict[str, list[TableRow]], list[TableRow]]":
        """
        Returns split doc table entry lists.
        First tuple element contains a dictionary with sub-key paths as key and
        list of TableEntry as value.
        Second tuple element contains a list of TableEntry for remaining
        (or non sub-key) documentation.
        """
        if not self.keys or max_level < _level:
            # This is an empty dict or
            # iteration limit reached
            return ({}, self.doc_table_rows)
        dict_entries = {}
        for key, obj in self.keys.items():
            split_doc, remain_doc = obj.get_doc_table_rows_dict(
                max_level=max_level, _level=_level + 1
            )
            if remain_doc:
                dict_entries[key] = self._format_key_table_entries(
                    entries=remain_doc, key=key
                )
            else:
                for sub_key_path, entries in split_doc.items():
                    sub_entries = [super().doc_table_rows[0]] + entries
                    dict_entries[f"{key}.{sub_key_path}"] = (
                        self._format_key_table_entries(entries=sub_entries, key=key)
                    )
        return (
            dict_entries,
            [],
        )

    def get_structured_doc_table_rows(
        self, structure: "dict|list[str]"
    ) -> "tuple[dict[str, list[TableRow]], list[TableRow]]":
        """
        Returns doc table entry lists according to given structure.
        First tuple element contains a dictionary with sub-key paths as key and
        list of TableEntry as value - sub-key paths matches the given structure.
        Second tuple element contains a list of TableEntry for remaining
        (not structure matching) documentation.
        """
        if not structure:
            return ({}, self.doc_table_rows)
        struct_doc = {}
        remain_doc = []
        for key, obj in self.keys.items():
            if key in structure:
                # key matches the given structure
                if isinstance(structure, list):
                    struct_doc[key] = self._format_key_table_entries(
                        entries=obj.doc_table_rows,
                        key=key,
                        add_top_row=self._schema.root is not self,
                    )
                elif isinstance(structure, dict):
                    sub_struct_doc, sub_remain_doc = obj.get_structured_doc_table_rows(
                        structure=structure[key]
                    )
                    if sub_remain_doc:
                        struct_doc[key] = self._format_key_table_entries(
                            entries=sub_remain_doc,
                            key=key,
                            add_top_row=self._schema.root is not self,
                        )
                    for key_path, entries in sub_struct_doc.items():
                        struct_doc[f"{key}.{key_path}"] = (
                            self._format_key_table_entries(
                                entries=entries,
                                key=key,
                                add_top_row=self._schema.root is not self,
                            )
                        )
                else:
                    raise ValueError(
                        f"Structure contains {type(structure)}. "
                        + "Only 'dict' and 'list' are supported."
                    )
            else:
                if not remain_doc and self._schema.root is not self:
                    remain_doc = [super().doc_table_rows[0]]
                remain_doc += self._format_key_table_entries(
                    entries=obj.doc_table_rows,
                    key=key,
                )
        return (
            struct_doc,
            remain_doc,
        )

    @property
    def doc_yaml(self) -> "list[str]":
        """
        Get data structure
        """
        if not self.keys:
            return ["{}"]
        ret_list = []
        for a_name, a_dt in self.keys.items():
            a_list = a_dt.doc_yaml
            if dict in a_dt.python_classes or list in a_dt.python_classes:
                # key is not a leaf
                if len(a_list) == 1 and a_list[0] == "{}":
                    ret_list.append(f"{a_name}: {a_list[0]}")
                else:
                    ret_list.append(f"{a_name}:")
                    ret_list += [f"  {x}" for x in a_list]
            else:
                ret_list.append(f"{a_name}: {a_list[0]}")
        return ret_list

    def _format_key_yaml_rows(self, rows: "list[str]", key: str) -> "list[str]":
        if len(rows) == 1 and (rows[0][0] == "<" or rows[0] == "{}"):
            # sub-key is not a structure data type or an empty dict
            return [f"{key}: {rows[0]}"]
        return [f"{key}:"] + [f"  {x}" for x in rows]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **keys** | <code>dict[dict&#124;str]</code> | | | |",
            "| &nbsp;&nbsp;< key_name > | <code>dict&#124;str</code> | | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 1* | <code>str</code> | | "
            + "Data type str (short definition) |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 2* | <code>dict/code> | | "
            + "Data type definition |",
            "| **restrictions** | <code>dict/code> | | | |",
            "| &nbsp;&nbsp;**required** | <code>list[str]&#124;str</code> | | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 1* | <code>str</code> | | Allowed_value: all | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;*option 2* | <code>list[str]</code> | | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;- < key_name > | <code>str</code> | | | |",
            "| &nbsp;&nbsp;**keys** | <code>dict/code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;**< key_name >** | <code>dict/code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;**excludes** | <code>list[str]/code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;**< key_name >** "
            + "| <code>str</code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;**requires** | <code>list[str]/code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;**< key_name >** "
            + "| <code>str</code>| | | |",
            "| &nbsp;&nbsp;**xor_required** | <code>list[list]/code>| | | |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;**-** | <code>list[str]/code>| | Min length: 2 | "
            + "Data's keys must match one and only one of these |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;**- < key_name >** | <code>str</code>| | | |",
            "| **ignore_unknown_keys** | <code>bool</code> "
            + "| <code>False</code><br><code>True</code>, if 'keys' is not defined | | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "",
            "keys:",
            "  <key_name>:",
            "    # Multitype!!!",
            "    <str>",
            "    # or",
            "    <dict>",
            "restrictions:",
            "  required:",
            "    # Multitype!!!",
            "    'all'",
            "    # or",
            "    - <str>",
            "  keys:",
            "    excludes:",
            "      - <key_name>",
            "    requires:",
            "      - <key_name>",
            "  xor_required:",
            "    -",
            "      - <key_name>",
            "ignore_unknown_keys: <true|false",
        ]


DOC_DT_CLASS = Dict  # pylint: disable=invalid-name



"""
Bool type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base

DOC_DT_NAME = "Bool"
DOC_DT_DESCRIPTION = """
A **bool** data type tests data for being an instance of
python's built-in class `bool`.
"""
DOC_DT_FEATURES = """
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = """
type: bool
description: "Example for a bool definition"
default_value: true

template_merge_options:
  recursive: < bool >
  list_merge: < str >
reference:
  key: ref_key
  mode: producer
  producer_scope: global
"""


class Bool(nadap.types.base.BaseType):
    """
    Bool datatype class
    """

    data_type_name = "bool"
    _cls_python_classes = [bool]


DOC_DT_CLASS = Bool  # pylint: disable=invalid-name



"""
Datatype base class
"""

# pylint: disable=too-few-public-methods

from nadap.base import UNDEFINED, SET_DEFAULTS, CONVERT_DATA, ValEnv, OPT
from nadap.errors import SchemaDefinitionError, DataValidationError, NadapReferenceError
import nadap.references
import nadap.results
import nadap.schema
from nadap.doc import TableRow, TextField


class BaseType:
    """
    Abstract root class for all data types
    """

    # pylint: disable=too-many-instance-attributes

    _cls_python_classes = []
    _support_ns_separator = False
    _support_replace_empty_to = False
    _convert_to_classes = {"str": str}
    _markdown_indent = "&nbsp;&nbsp;"
    data_type_name = None
    _doc_data_type_name = None

    def __init__(
        self,
        schema: "nadap.schema.Schema",
        definition: dict,
        schema_path: str,
    ):
        self.default_value = UNDEFINED
        self.description = ""
        self.references = []
        self.python_classes = self._cls_python_classes[:]
        self._convert_to = None
        self._doc_value_name = None
        self._schema = schema

        self._pop_options(definition=definition, schema_path=schema_path)
        nadap.schema.no_more_definition_options(
            definition=definition, source="data type", path=schema_path
        )
        self._validate_options(schema_path=schema_path)

    @staticmethod
    def _raise_exceptions(messages: "list[str]", path: str, env: "ValEnv|None"):
        if messages:
            if isinstance(messages, str):
                messages = [messages]
            if env is None:
                raise SchemaDefinitionError(
                    messages[0],
                    path,
                )
            for m in messages:
                env.findings.append(
                    nadap.results.ValidationFinding(
                        message=m,
                        path=path,
                    )
                )
            raise DataValidationError()

    @property
    def _data_type_mismatch_message(self) -> str:
        """
        Get error message if data does not match data type
        """
        return f"Data is not an instance of {self._python_classes_list_str}"

    @staticmethod
    def _refopt_mode(mode: str, schema_path: str) -> "OPT":
        if mode == "unique":
            return OPT.UNIQUE | OPT.PRODUCER
        if mode == "producer":
            return OPT.PRODUCER
        if mode == "consumer":
            return OPT.CONSUMER
        raise SchemaDefinitionError("Unsupported mode", f"{schema_path}.mode")

    def _pop_reference_definition(self, ref_definition: dict, schema_path: str):
        # pylint: disable=too-many-branches,unsupported-membership-test
        nadap.schema.is_dict(ref_definition, schema_path)
        opt_path = f"{schema_path}."

        if not ref_definition:
            return
        # parse "key"
        if "key" not in ref_definition.keys():
            raise SchemaDefinitionError("Key missing", f"{opt_path}key")
        ref_key = ref_definition.pop("key")
        nadap.schema.is_str(ref_key, f"{opt_path}key")

        # Parse "mode"
        ref_options = self._refopt_mode(
            ref_definition.pop("mode", "unique"), schema_path
        )
        u_scope = ref_definition.pop("unique_scope", "global")
        p_scope = ref_definition.pop("producer_scope", "global")
        c_scope = ref_definition.pop("consumer_scope", "global")
        orphan = ref_definition.pop("allow_orphan_producer", True)
        ref_credits = ref_definition.pop("credits", None)
        if OPT.UNIQUE in ref_options:
            if u_scope == "global":
                ref_options = ref_options | OPT.UNIQUE_GLOBAL
            elif u_scope != "namespace":
                raise SchemaDefinitionError(
                    "Must be either 'global' or 'namespace'",
                    opt_path + "unique_scope",
                )
            if ref_credits is not None and ref_credits < 0:
                raise SchemaDefinitionError(
                    "Must be >=0",
                    opt_path + "credits",
                )
        if OPT.PRODUCER in ref_options:
            if p_scope == "global":
                ref_options = ref_options | OPT.PRODUCER_GLOBAL
            elif p_scope != "namespace":
                raise SchemaDefinitionError(
                    "Must be either 'global' or 'namespace'",
                    opt_path + "producer_scope",
                )
            if orphan:
                ref_options = ref_options | OPT.ALLOW_ORPHAN_PRODUCER
            if ref_credits is not None and ref_credits < 0:
                raise SchemaDefinitionError(
                    "Must be >=0",
                    opt_path + "credits",
                )
        if OPT.CONSUMER in ref_options:
            if c_scope == "global":
                ref_options = ref_options | OPT.CONSUMER_GLOBAL
            elif c_scope != "namespace":
                raise SchemaDefinitionError(
                    "Must be either 'global' or 'namespace'",
                    opt_path + "consumer_scope",
                )
            if ref_credits is not None and ref_credits < 0:
                raise SchemaDefinitionError(
                    "Must be >0",
                    opt_path + "credits",
                )

        if not isinstance(orphan, bool):
            raise SchemaDefinitionError(
                "Must be bool", opt_path + "allow_orphan_producer"
            )

        if self._support_ns_separator:
            ns_sep = ref_definition.pop("namespace_separator_char", None)
            if ns_sep is not None and (not isinstance(ns_sep, str) or len(ns_sep) != 1):
                raise SchemaDefinitionError(
                    "Must be a single character string",
                    f"{schema_path}.namespace_separator_char",
                )
        else:
            ns_sep = None

        nadap.schema.no_more_definition_options(
            definition=ref_definition, source="reference", path=schema_path
        )

        self.references.append(
            nadap.references.RefDef(
                ref_key=ref_key,
                ref_options=ref_options,
                ref_credits=ref_credits,
                ns_separator=ns_sep,
            )
        )

    def _pop_reference(self, ref_definition: "dict|str", schema_path: str):
        # If reference is a string, bypass parsing reference options:
        ref_path = f"{schema_path}.reference"
        if isinstance(ref_definition, str):
            ref_definition = {"key": ref_definition}
        self._pop_reference_definition(ref_definition, ref_path)

    def _pop_references(self, ref_definitions: dict | str, schema_path: str):
        ref_path = f"{schema_path}.references"
        nadap.schema.is_list(ref_definitions, ref_path)

        for index, ref_definition in enumerate(ref_definitions):
            self._pop_reference_definition(ref_definition, f"{ref_path}[{index}]")

    def _pop_convert_options(self, definition: dict, schema_path: str):
        con_to = "convert_to"
        if con_to in definition:
            c_path = f"{schema_path}.{con_to}"
            if not self._convert_to_classes:
                raise SchemaDefinitionError(
                    "Data type does not support conversion", c_path
                )
            target_class = definition.pop(con_to)
            nadap.schema.is_str(target_class, c_path)
            if target_class not in self._convert_to_classes:
                raise SchemaDefinitionError("Target data type not supported", c_path)
            self._convert_to = self._convert_to_classes[target_class]

    def _pop_options(self, definition: dict, schema_path: str):
        """
        Pop data type options from definition
        """
        self.default_value = definition.pop("default_value", UNDEFINED)

        self.description = definition.pop("description", self.description)
        self._pop_convert_options(definition=definition, schema_path=schema_path)

        dvn = "doc_value_name"
        if dvn in definition:
            self._doc_value_name = definition.pop(dvn)
            nadap.schema.is_str(self._doc_value_name, f"{schema_path}.{dvn}")

        if "reference" in definition and "references" in definition:
            raise SchemaDefinitionError(
                "Data type option 'references' excludes option 'reference'",
                schema_path,
            )
        self._pop_reference(definition.pop("reference", {}), schema_path)
        self._pop_references(definition.pop("references", []), schema_path)

    def _validate_options(self, schema_path: str):
        if self.default_value is not UNDEFINED:
            self._test_data_type(self.default_value, f"{schema_path}.default_value")
        nadap.schema.is_str(self.description, f"{schema_path}.description")

    def _set_default(self, data: any):
        """
        Set values of nested data types to default if missing.
        """
        return data

    def _detect_data_format(self, data) -> str:  # pylint: disable=unused-argument
        return None

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        if fmt is None:
            return data
        return data

    def _convert_data(self, data: any) -> any:
        """
        Change data type
        """
        if self._convert_to is not None:
            return self._convert_to(data)
        return data

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if SET_DEFAULTS in env.flags:
            data = self._set_default(data)
        if CONVERT_DATA in env.flags:
            data = self._convert_data(data)
        return data

    def _convert_reference_data(self, data: any) -> any:
        """
        Convert data to reference data
        """
        return data

    def validate(
        self,
        data: any,
        path: str,
        env: "ValEnv",
    ) -> "any":
        """
        Check if data matches defined data type,
        apply preprocessing and
        complies to referencing
        """
        data = self._validate_data(data=data, path=path, env=env)
        data = self._preprocess_data(data=data, env=env)
        return self._reference_data(data=data, path=path, env=env)

    def _validate_data(self, data: any, path: str, env: ValEnv) -> "any":
        """
        Check if data matches defined data type.
        """
        return self._test_data_type(data=data, path=path, env=env)

    def _reference_data(self, data: any, path: str, env: ValEnv) -> "any":
        """
        Check if data matches defined data type.
        """
        for ref_def in self.references:
            if ref_def.ns_separator and ref_def.ns_separator in data:
                provider_ns, value = data.split(ref_def.ns_separator, 1)
                element = nadap.references.ConsumerElement(
                    ref_def=ref_def,
                    path=path,
                    value=self._convert_reference_data(value),
                    provider_namespace=provider_ns,
                )
            else:
                element = nadap.references.ReferenceElement(
                    ref_def=ref_def,
                    path=path,
                    value=self._convert_reference_data(data),
                )
            try:
                env.references.add_element(element)
            except NadapReferenceError as e:
                env.findings.append(
                    nadap.references.ReferenceFinding(message=str(e), path=path)
                )
        return data

    def _test_data_type(self, data: any, path: str, env: "ValEnv" = None) -> "any":
        if not self.python_classes:
            return data
        if not isinstance(data, tuple(self.python_classes)):
            if env is None:
                raise SchemaDefinitionError(
                    self._data_type_mismatch_message,
                    path,
                )
            self._create_finding_with_error(self._data_type_mismatch_message, path, env)
        return data

    def _create_finding_with_error(self, msg: str, path: str, env: "ValEnv"):
        env.findings.append(
            nadap.results.ValidationFinding(
                message=msg,
                path=path,
            )
        )
        raise DataValidationError()

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        return TextField()

    @property
    def doc_value_name(self) -> str:
        """
        Get Value name for documentation
        """
        if self._doc_value_name is not None:
            return self._doc_value_name
        if self.python_classes:
            return "|".join(sorted([x.__name__ for x in self.python_classes]))
        return self.data_type_name

    @property
    def _python_classes_list_str(self) -> str:
        if len(self.python_classes) == 1:
            return f"'{self.python_classes[0].__name__}'"
        cls_str = sorted([f"'{x.__name__}'" for x in self.python_classes])
        return f"{', '.join(cls_str[:-1])} or {cls_str[-1]}"

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        if self.python_classes:
            return "|".join(sorted([x.__name__ for x in self.python_classes]))
        return f"{self.data_type_name}"

    @property
    def doc_types(self) -> list[str]:
        """Get list of data type strings"""
        if self.python_classes:
            return sorted([x.__name__ for x in self.python_classes])
        return [self.data_type_name]

    @property
    def doc_table_rows(self) -> "list[TableRow]":
        """
        Get list of documentation table entries
        """
        return [
            TableRow(
                type=self.doc_types,
                default=self.default_value,
                restrictions=self.restrictions,
                description=self.description.splitlines(),
            )
        ]

    def get_doc_table_rows_dict(
        self, max_level: int = 0, _level: int = 0
    ) -> "tuple[dict[str, list[TableRow]], list[TableRow]]":
        """
        Returns split doc table entry lists.
        First tuple element contains a dictionary with sub-key paths as key and
        list of TableEntry as value.
        Second tuple element contains a list of TableEntry for remaining
        (or non sub-key) documentation.
        """
        # pylint: disable=unused-argument
        return (
            {},
            self.doc_table_rows,
        )

    def get_structured_doc_table_rows(
        self, structure: "dict|list[str]"
    ) -> "tuple[dict[str, list[TableRow]], list[TableRow]]":
        """
        Returns doc table entry lists according to given structure.
        First tuple element contains a dictionary with sub-key paths as key and
        list of TableEntry as value - sub-key paths matches the given structure.
        Second tuple element contains a list of TableEntry for remaining
        (not structure matching) documentation.
        """
        # pylint: disable=unused-argument
        return (
            {},
            self.doc_table_rows,
        )

    @classmethod
    def _doc_md_type(cls):
        if cls._cls_python_classes:
            return (
                "<code>"
                + "&#124;".join(sorted([x.__name__ for x in cls._cls_python_classes]))
                + "</code>"
            )
        return "`any`"

    @classmethod
    def _doc_yaml_type(cls):
        if cls._cls_python_classes:
            return "|".join(sorted([x.__name__ for x in cls._cls_python_classes]))
        return "any"

    @property
    def doc_yaml(self) -> "any":
        """
        Get data structure
        """
        return [f"<{self.yaml_data_type}>"]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        ret_list = [
            f"| **type** | <code>str</code> | | '{cls.data_type_name}' | |",
            "| **description** | <code>str</code> | | | Data type description |",
            "| **default_value** | | | Must match data type's type(s) | "
            + "Value used to set as default value |",
        ]
        if cls._convert_to_classes:
            ret_list.append(
                "| **convert_to** | <code>str</code> | | Allowed values:<br>- "
                + "<br>- ".join(sorted(cls._convert_to_classes.keys()))
                + " | |"
            )
        return ret_list

    @classmethod
    def _doc_ref_options_md_table_rows(cls, indent: int = 0) -> list[str]:
        ret_list = [
            f"| {'&nbsp;&nbsp;' * indent}**key** | <code>str</code> | | required | |",
            f"| {'&nbsp;&nbsp;' * indent}**mode** | <code>str</code> | "
            + "| Allowed values:<br>- unique<br>- "
            + "producer<br>- consumer | |",
            f"| {'&nbsp;&nbsp;' * indent}**unique_scope** "
            + "| <code>str</code> | | Allowed values:<br>- "
            + "global<br>- namespace | |",
            f"| {'&nbsp;&nbsp;' * indent}**producer_scope** "
            + "| <code>str</code> | | Allowed values:<br>- "
            + "global<br>- namespace | |",
            f"| {'&nbsp;&nbsp;' * indent}**consumer_scope** "
            + "| <code>str</code> | | Allowed values:<br>- "
            + "global<br>- namespace | |",
            f"| {'&nbsp;&nbsp;' * indent}**allow_orphan_producer** "
            + "| <code>bool</code> | <code>True</code> | | |",
            f"| {'&nbsp;&nbsp;' * indent}**credit** | <code>int</code> "
            + "| 1 (if mode consumer) | min: 0 | |",
        ]
        if cls._support_ns_separator:
            ret_list.append(
                f"| {'&nbsp;&nbsp;' * indent}**namespace_separator_char** "
                + "| <code>str</code> | | length must be 1 | |"
            )
        return ret_list

    @classmethod
    def _doc_options_md_lower_part(cls) -> list[str]:
        ret_list = [
            "| **template** | <code>str</code>| | must be definded in schema "
            + "| Merge options from this "
            + "template |",
            "| **template_merge_options** | <code>dict</code>| | "
            + "| Define schema-global template merge "
            + "instructions |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;**recursive** | <code>bool</code> "
            + "| <code>True</code> | | Merge data type "
            + "definition options in template options recursively.<br>If false, data type "
            + "defintion options overwrites template options. |",
            "| &nbsp;&nbsp;&nbsp;&nbsp;**list_merge** | <code>str</code> | append_rp | "
            + "Allowed values:<br>- append<br>- append_rp<br>- prepend<br>- prepend_rp<br>- "
            + "replace | How lists within options data should be merged. |",
            "| **reference** | <code>dict&#124;str</code> | | | |",
            "| &nbsp;&nbsp;*option 1* | <code>str</code> | | "
            + "| Reference keys (implies unique global mode) |",
            "| &nbsp;&nbsp;*option 2* | <code>dict</code> | | | |",
        ]
        ret_list.extend(cls._doc_ref_options_md_table_rows(indent=1))
        ret_list.extend(
            [
                "| **references** | <code>list[dict&#124;str]</code> | | | |",
                "| &nbsp;&nbsp;- <reference defintiion> | <code>dict&#124;str</code> | | | |",
                "| &nbsp;&nbsp;&nbsp;&nbsp;option 1 | <code>str</code> | | | Reference keys "
                + "(implies unique global mode) |",
                "| &nbsp;&nbsp;&nbsp;&nbsp;option 2 | <code>dict</code> | | | |",
            ]
        )
        ret_list.extend(cls._doc_ref_options_md_table_rows(indent=2))
        ret_list.append(
            "| **doc_value_name** | <code>str</code> | | | Set value name for documentation |"
        )
        return ret_list

    @classmethod
    def doc_options_md_table_rows(cls) -> list[str]:
        """
        Get a list of markdown table rows with all
        definition options for this data type for documentation
        """
        return cls._doc_options_md_upper_part() + cls._doc_options_md_lower_part()

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        ret_list = [
            f"type: {cls.data_type_name}",
            "description: <str>",
            "default_value: <default_value>",
            "",
        ]
        if cls._convert_to_classes:
            ret_list.append(
                f"convert_to: <{'|'.join(sorted(cls._convert_to_classes.keys()))}>"
            )
        return ret_list

    @classmethod
    def _doc_ref_options_yaml_rows(cls, indent: int = 0) -> list[str]:
        ret_list = [
            f"{'  ' * indent}key: <str>",
            f"{'  ' * indent}mode: <unique|producer|consumer>",
            f"{'  ' * indent}unique_scope: <global|namespace>",
            f"{'  ' * indent}producer_scope: <global|namespace>",
            f"{'  ' * indent}consumer_scope: <global|namespace>",
            f"{'  ' * indent}allow_orphan_producer: <true|false>",
            f"{'  ' * indent}credits: <int>",
        ]
        if cls._support_ns_separator:
            ret_list.append(f"{'  ' * indent}namespace_separator_char: <str>")
        return ret_list

    @classmethod
    def _doc_options_yaml_lower_part(cls) -> list[str]:
        ret_list = [
            "",
            "template: <str>",
            "template_merge_options:",
            "  recursive: <true|false>",
            "  list_merge: <append|append_rp|prepend|prepend_rp|replace>",
            "",
            "reference: <reference_definition>",
            "  # Multitype!!!",
            "  <str>",
            "  # or:",
        ]
        ret_list.extend(cls._doc_ref_options_yaml_rows(indent=1))
        ret_list.extend(
            [
                "references:",
                "  - <reference_definition>",
                "    # Multitype!!!",
                "    <str>",
                "    # or",
            ]
        )
        ret_list.extend(cls._doc_ref_options_yaml_rows(indent=2))
        ret_list.append("")
        ret_list.append("doc_value_name: <str>")
        return ret_list

    @classmethod
    def doc_options_yaml_rows(cls) -> list[str]:
        """
        Get a list of YAML text rows with all
        definition options for this data type for documentation
        """
        return cls._doc_options_yaml_upper_part() + cls._doc_options_yaml_lower_part()



"""
NotAllowedValueMixin
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.schema
from nadap.base import ValEnv, number_to_str_number
from nadap.doc import UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField


class NotAllowedValueMixin:
    """
    Add not allowed value tests for data
    """

    def __init__(self, **kwargs):
        self.not_allowed_values = None
        self._not_allowed_value_formats = []
        super().__init__(**kwargs)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        """
        Check if data doesn't match not allowed values
        """
        data = super()._validate_data(data=data, path=path, env=env)
        if self.not_allowed_values is not None and data in self.not_allowed_values:
            self._create_finding_with_error(
                msg="Data is a not allowed value",
                path=path,
                env=env,
            )
        return data

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        if self.not_allowed_values is not None:
            nadap.schema.is_list(
                self.not_allowed_values, f"{schema_path}.not_allowed_values"
            )
            values = []
            for index, v in enumerate(self.not_allowed_values):
                self._not_allowed_value_formats.append(self._detect_data_format(v))
                values.append(
                    self._test_data_type(
                        data=v, path=f"{schema_path}.not_allowed_values[{index}]"
                    )
                )
            self.not_allowed_values = values

    def _pop_options(self, definition: dict, schema_path: str):
        self.not_allowed_values = definition.pop(
            "not_allowed_values", self.not_allowed_values
        )
        super()._pop_options(definition=definition, schema_path=schema_path)

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if self.not_allowed_values:
            tf.append("not allowed values:")
            out_values = []
            for index, v in enumerate(self.not_allowed_values):
                out_values.append(
                    self._apply_data_format(v, self._not_allowed_value_formats[index])
                )
            tf.append(
                UnorderedTextList([str(number_to_str_number(x)) for x in out_values])
            )
        return tf

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **not_allowed_values** | <code>list</code> | | | "
            + " Data mustn't match all of these values |",
            f"| &nbsp;&nbsp;- < value > | {cls._doc_md_type()}"
            + " | | Must match data type's type(s) | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "not_allowed_values:",
            f"  - <{cls._doc_yaml_type()}>",
        ]



"""
Accuracy Mixin
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING

import nadap.schema
from nadap.base import ValEnv
from nadap.errors import SchemaDefinitionError, DataValidationError

if TYPE_CHECKING:
    from nadap.doc import TextField


class AccuracyMixin:
    """
    Add options for maximum_decimals
    """

    def __init__(self, **kwargs):
        self._maximum_decimals = None
        self._round_to = None
        super().__init__(**kwargs)

    def _pop_options(self, definition: dict, schema_path: str):
        self._maximum_decimals = definition.pop("maximum_decimals", None)
        self._round_to = definition.pop("round_to_decimals", None)
        super()._pop_options(definition=definition, schema_path=schema_path)

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)
        if self._maximum_decimals is not None:
            _path = f"{schema_path}.maximum_decimals"
            nadap.schema.is_int(self._maximum_decimals, _path)
            if not 0 <= self._maximum_decimals <= 17:
                raise SchemaDefinitionError("Must be >= 0 and <= 17", _path)
        if self._round_to is not None:
            _path = f"{schema_path}.round_to_decimals"
            nadap.schema.is_int(self._round_to, _path)
            if not 0 <= self._round_to <= 17:
                raise SchemaDefinitionError("Must be >= 0 and <= 17", _path)

    def _validate_data(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> "any":
        data = super()._validate_data(data=data, path=path, env=env)
        if self._maximum_decimals is not None and data != round(
            data, self._maximum_decimals
        ):
            env.findings.append(
                nadap.results.ValidationFinding(
                    "Float value has too many decimals",
                    path,
                )
            )
            raise DataValidationError()
        return data

    def _convert_data(self, data: any) -> any:
        if self._round_to is not None:
            data = round(data, self._round_to)
        return super()._convert_data(data)

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        if self._maximum_decimals is not None:
            tf.append(f"max allowed decimals: {self._maximum_decimals}")
        return tf

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        return super()._doc_options_md_upper_part() + [
            "| **maximum_decimals** | <code>int</code> | | Must be >= 0 and <= 17 | "
            + "Value's accuracy must be lower or equal |",
            "| **round_to_decimals** | <code>int</code> | | Must be >= 0 and <= 17 | "
            + "Convert value to this accuracy |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            "maximum_decimals: <int>",
            "round_to_decimals: <int>",
        ]



"""
Class Schema
"""

# pylint: disable=too-few-public-methods

import copy
import re
import nadap.types.any
import nadap.types.base
import nadap.types.bgp_as
import nadap.types.bgp_community
import nadap.types.bgp_ext_community
import nadap.types.bool
import nadap.types.bool_false
import nadap.types.bool_true
import nadap.types.byte4_value
import nadap.types.dict
import nadap.types.enum
import nadap.types.float
import nadap.types.hostname_or_ip
import nadap.types.hostname
import nadap.types.idlist
import nadap.types.int
import nadap.types.int8
import nadap.types.int16
import nadap.types.int32
import nadap.types.ip_address
import nadap.types.ip4_address
import nadap.types.ip6_address
import nadap.types.ip_interface
import nadap.types.ip4_interface
import nadap.types.ip6_interface
import nadap.types.ip_network
import nadap.types.ip4_network
import nadap.types.ip6_network
import nadap.types.list
import nadap.types.mac_address
import nadap.types.multitype
import nadap.types.multitype2
import nadap.types.none
import nadap.types.number
import nadap.types.str
import nadap.types.str_float
import nadap.types.str_int
import nadap.types.str_number
import nadap.types.uint8
import nadap.types.uint16
import nadap.types.uint32
from nadap.doc import Doc
import nadap.references
from nadap.base import merge_dictionaries, RegexObject
from nadap.errors import SchemaDefinitionError

if hasattr(re, "PatternError"):
    PatternError = re.PatternError
else:
    PatternError = re.error


def is_list(l, path: str):
    """
    Test if given value (t) is a instance of class list; else raise an error.
    """
    if not isinstance(l, list):
        raise SchemaDefinitionError(msg="Must be a list", path=path)


def is_non_empty_list(l, path: str):
    """
    Test if given value (t) is a instance of class list
    and is not empty; else raise an error.
    """
    is_list(l, path)
    if not l:
        raise SchemaDefinitionError(
            msg="Mustn't be empty",
            path=path,
        )


def is_dict(d, path: str):
    """
    Test if given value (d) is a instance of class dict; else raise an error.
    """
    if not isinstance(d, dict):
        raise SchemaDefinitionError(msg="Must be a dictionary", path=path)


def is_str(s, path: str):
    """
    Test if given value (s) is a instance of class str; else raise an error.
    """
    if not isinstance(s, str):
        raise SchemaDefinitionError(msg="Must be a string", path=path)


def is_bool(b, path: str):
    """
    Test if given value (b) is a instance of class bool; else raise an error.
    """
    if not isinstance(b, bool):
        raise SchemaDefinitionError(msg="Must be a boolean", path=path)


def is_int(s, path: str):
    """
    Test if given value (s) is a instance of class int; else raise an error.
    """
    if not isinstance(s, int):
        raise SchemaDefinitionError(msg="Must be an integer", path=path)


def is_list_merge_option(o: str, path: str):
    """
    Test if given list_merge option is valid
    """
    is_str(o, path)
    if o not in ["append", "append_rp", "prepend", "prepend_rp", "replace"]:
        raise SchemaDefinitionError(msg="Invalid list merge option", path=path)


def no_more_definition_options(definition: dict, source: str, path: str):
    """
    Test if given definition dictionary is empty.
    Else raise an error for containing (unknown) options.
    """
    if definition:
        str_list = list(definition)
        if len(definition) > 1:
            last = str_list[-1]
            str_list = str_list[:-1]
            msg = f"Unknown {source} options {', '.join(str_list)} and {last}"
        else:
            msg = f"Unknown {source} option {str_list[0]}"
        raise SchemaDefinitionError(msg, path)


def compile_regex_string(
    pattern: str, multiline: bool, fullmatch: bool, schema_path: str
) -> "RegexObject":
    """
    Try to compile regex object based on given arguments
    """
    try:
        re_obj = RegexObject(pattern, multiline, fullmatch)
    except (PatternError, TypeError) as e:
        raise SchemaDefinitionError("Not a valid regex pattern", schema_path) from e
    if not pattern:
        raise SchemaDefinitionError(
            "Empty string not allowed as regex pattern", schema_path
        )
    return re_obj


def match_regex_objects(regex_objects: list, data: str) -> tuple[bool, str]:
    """
    Check if string matches any of the regex objects
    """
    for re_obj in regex_objects:
        if re_obj.match(data):
            return True, re_obj.pattern
    return False, ""


class Schema:
    """
    Class Schema
    """

    # pylint: disable=too-many-instance-attributes

    type_class_matching = {
        "4byte_value": nadap.types.byte4_value.Byte4Value,
        "any": nadap.types.any.Any,
        "bgp_as": nadap.types.bgp_as.BgpAs,
        "bgp_community": nadap.types.bgp_community.BgpCommunity,
        "bgp_ext_community": nadap.types.bgp_ext_community.BgpExtCommunity,
        "bool": nadap.types.bool.Bool,
        "bool_false": nadap.types.bool_false.BoolFalse,
        "bool_true": nadap.types.bool_true.BoolTrue,
        "dict": nadap.types.dict.Dict,
        "enum": nadap.types.enum.Enum,
        "float": nadap.types.float.Float,
        "hostname": nadap.types.hostname.Hostname,
        "hostname_or_ip": nadap.types.hostname_or_ip.HostnameOrIP,
        "idlist": nadap.types.idlist.IDList,
        "int": nadap.types.int.Int,
        "int8": nadap.types.int8.Int8,
        "int16": nadap.types.int16.Int16,
        "int32": nadap.types.int32.Int32,
        "ip_address": nadap.types.ip_address.IpAddress,
        "ip4_address": nadap.types.ip4_address.Ip4Address,
        "ip6_address": nadap.types.ip6_address.Ip6Address,
        "ip_network": nadap.types.ip_network.IpNetwork,
        "ip4_network": nadap.types.ip4_network.Ip4Network,
        "ip6_network": nadap.types.ip6_network.Ip6Network,
        "ip_interface": nadap.types.ip_interface.IpInterface,
        "ip4_interface": nadap.types.ip4_interface.Ip4Interface,
        "ip6_interface": nadap.types.ip6_interface.Ip6Interface,
        "list": nadap.types.list.List,
        "multitype": nadap.types.multitype.MultiType,
        "multitype2": nadap.types.multitype2.MultiType2,
        "mac_address": nadap.types.mac_address.MacAddress,
        "none": nadap.types.none.Null,
        "number": nadap.types.number.Number,
        "str": nadap.types.str.Str,
        "str_float": nadap.types.str_float.StrFloat,
        "str_int": nadap.types.str_int.StrInt,
        "str_number": nadap.types.str_number.StrNumber,
        "uint8": nadap.types.uint8.Uint8,
        "uint16": nadap.types.uint16.Uint16,
        "uint32": nadap.types.uint32.Uint32,
    }

    def __init__(self, definition: dict = None):
        self.references = nadap.references.References()
        self.reference_findings = []

        self.name = ""
        self.description = ""

        self.root = None
        self.custom_data_types = {}
        self.templates = {}
        self.template_merge_recursive = True
        self.template_list_merge = "append_rp"
        self.template_stack = []

        if definition is not None:
            self.load_definition(definition)

    def _is_loaded(self):
        """
        Check if Schema has loaded a schema definition
        """
        if self.root is None:
            raise RuntimeError("Schema definition hasn't been loaded")

    def _merge_template_definition(self, definition: dict, path: str) -> dict:
        definition = copy.deepcopy(definition)
        template_name = definition.pop("template", None)
        if template_name:
            if f"template.{template_name}" in path.split(" > "):
                raise SchemaDefinitionError(
                    "Nested template loop", f"{path} > template.{template_name}"
                )
            if template_name not in self.templates:
                raise SchemaDefinitionError(
                    f"Unknown template {template_name}",
                    f"{path}.template",
                )
            t_merge_opts = definition.pop("template_merge_options", {})
            list_merge_option = t_merge_opts.pop("list_merge", self.template_list_merge)
            is_list_merge_option(
                list_merge_option, f"{path}.template_merge_options.list_merge"
            )
            definition = merge_dictionaries(
                left=self.templates[template_name],
                right=definition,
                recursive=t_merge_opts.pop("recursive", self.template_merge_recursive),
                list_merge=list_merge_option,
            )
            definition = self._merge_template_definition(
                definition, f"{path} > template.{template_name}"
            )
        return definition

    def load_data_type_by_definition(
        self, definition: dict | str, path: str
    ) -> nadap.types.base.BaseType:
        """
        Identify and load data type object by given definition.
        """
        if isinstance(definition, str):
            definition = {"type": definition}
        elif not isinstance(definition, dict):
            raise SchemaDefinitionError(
                "Data type definition must be a string or dictionary", path
            )
        # First look up if this is an custom data type:
        if cdt_def := self.custom_data_types.get(definition.get("type", None), None):
            new_def = copy.deepcopy(cdt_def)
            new_def.update(definition)
            definition = new_def
        if template_name := definition.pop("template", None):
            if template_name not in self.templates:
                raise SchemaDefinitionError(
                    f"Unknown template {template_name}",
                    f"{path}.template",
                )
            t_merge_opts = definition.pop("template_merge_options", {})
            list_merge_option = t_merge_opts.pop("list_merge", self.template_list_merge)
            is_list_merge_option(
                list_merge_option, f"{path}.template_merge_options.list_merge"
            )
            definition = merge_dictionaries(
                left=self.templates[template_name],
                right=definition,
                recursive=t_merge_opts.pop("recursive", self.template_merge_recursive),
                list_merge=list_merge_option,
            )
        _type = definition.pop("type", None)
        if _type is None:
            raise SchemaDefinitionError(msg="Key 'type' missing.", path=path)
        # If this is a CDT options have already been merged,
        # but "type" was overwritten. Pull in "type" excplicitly from CDT.
        if _type in self.custom_data_types:
            _type = self.custom_data_types[_type]["type"]
        if _type not in self.type_class_matching:
            raise SchemaDefinitionError(
                msg=f"Data type '{_type}' not found",
                path=path,
            )
        data_type_class = self.type_class_matching[_type]
        data_type = data_type_class(
            schema=self, definition=definition, schema_path=path
        )
        if unknown_options := sorted(definition):
            if len(unknown_options) > 1:
                unknown_options = [f"'{x}'" for x in unknown_options]
                msg = f"Unknown options {', '.join(unknown_options)}"
            else:
                msg = f"Unknown option '{unknown_options[0]}'"
            raise SchemaDefinitionError(
                msg=f"{msg} for data type {data_type_class.data_type_name}", path=path
            )
        return data_type

    def _parse_custom_data_types(self, definition):
        for name, cdt_definition in definition.pop("custom_data_types", {}).items():
            if cdt_definition.get("type", None) not in self.type_class_matching:
                raise SchemaDefinitionError(
                    "Not a built-in data type", f"custom_data_types.{name}"
                )
            self.load_data_type_by_definition(
                copy.deepcopy(cdt_definition),
                f"custom_data_types.{name}",
            )
            self.custom_data_types[name] = cdt_definition

    def _parse_template_merge_options(self, definition):
        template_merge_options = definition.pop("template_merge_options", {})
        path = "template_merge_options"
        is_dict(template_merge_options, path)
        self.template_merge_recursive = template_merge_options.pop(
            "recursive", self.template_merge_recursive
        )
        self.template_list_merge = template_merge_options.pop(
            "list_merge", self.template_list_merge
        )
        is_list_merge_option(self.template_list_merge, f"{path}.list_merge")

    def _parse_templates(self, definition):
        self.templates = definition.pop("templates", {})
        # Check if template definition is a correct data type definition.
        for name, t_definition in self.templates.items():
            self.templates[name] = self._merge_template_definition(
                t_definition, f"template.{name}"
            )

    def load_definition(self, definition: dict) -> "nadap.types.base.DataType":
        """
        Parse schema definition and returns the root data type of this schema.
        """
        definition = copy.deepcopy(definition)
        self.name = definition.pop("name", "")
        self.description = definition.pop("description", "")
        self._parse_template_merge_options(definition)
        self._parse_templates(definition)
        self._parse_custom_data_types(definition)
        try:
            root = definition.pop("root")
        except KeyError as e:
            raise SchemaDefinitionError(f"Missing schema definition key {e}") from e
        if unknown_options := sorted(definition):
            if len(unknown_options) > 1:
                unknown_options = [f"'{x}'" for x in unknown_options]
                msg = f"Unknown schema options {', '.join(unknown_options)}"
            else:
                msg = f"Unknown schema option '{unknown_options[0]}'"
            raise SchemaDefinitionError(msg=msg)
        self.root = self.load_data_type_by_definition(
            definition=root,
            path="root",
        )
        return self.root

    @property
    def _doc_table(self) -> "Doc":
        entries = self.root.doc_table_rows
        if (
            len(entries) > 1
            and (entries[0].type.types in [["dict"], ["list"]])
            and not entries[0].default
            and not entries[0].description
        ):
            entries = entries[1:]
        return Doc(entries)

    def doc(self, fmt: str) -> str:
        """
        Get documentation for this data schema

        Arguments:
            format:     str
                        Intended output format
                        Allowed values:
                        - markdown
                        - yaml
                        - none
        Returns:        str
                        Documentation in requested format
        """
        self._is_loaded()

        doc_table = self._doc_table

        if fmt == "markdown":
            ret_str = doc_table.markdown_table
        elif fmt == "yaml":
            ret_str = doc_table.nice_yaml
        elif fmt == "none":
            ret_str = doc_table
        else:
            raise RuntimeError("Unknown documentation format")
        return ret_str

    def doc_sections(self, fmt: str, level: int = 0) -> dict:
        """
        Get documentation split into sections.
        If data type is an object data type, documentation is
        split into each key. This is done recursively up to 'level' deep.

        Arguments:
            format:     str
                        Intended output format
                        Allowed values:
                        - markdown
                        - yaml
                        - none
            level:      int >= 0
                        Split level
        Returns:        dict
                        Keys: keys in data type
                        Values:
                            str - Key path within data schema
                            str - Documentation of sub data schema
        """
        # Check if schema is initialized
        self._is_loaded()

        # Parse split rows to intended format
        dict_str = {}
        split_doc, remain_doc = self.root.get_doc_table_rows_dict(max_level=level)
        if remain_doc:
            raise RuntimeError("Schema's root data type is not 'dict'")
        if fmt == "markdown":
            for key_path, entries in split_doc.items():
                dict_str[key_path] = Doc(entries).markdown_table
        elif fmt == "yaml":
            for key_path, entries in split_doc.items():
                dict_str[key_path] = Doc(entries).nice_yaml
        elif fmt == "none":
            for key_path, entries in split_doc.items():
                dict_str[key_path] = Doc(entries)
        else:
            raise RuntimeError("Unknown documentation format")
        return dict_str

    def structured_doc(
        self, fmt: str, structure: "dict|list[str]"
    ) -> "tuple[dict[str, str], str]":
        """
        Get documentation snippets based on given structure.

        Structure means a list of strings or lists of strings
        in a nested dictionary.
        Dictionaries in the structure shall map to
        the 'dict' data types in the schema.
        List of strings shall represent keys in the 'dict' data type.

        Matching dict/keys will be returned as dictioniary with key path as key
        and documentation string as value. (First tuple element)

        Non-matching keys or structure elements will be return in
        a documentation string. (Second tuple element)

        Arguments:
            format:     str
                        Intended output format
                        Allowed values:
                        - markdown
                        - yaml
                        - none
            structure:  List of strings (keys)
                        or nested dicts of list of strings
        """
        # Check if schema is initialized
        self._is_loaded()

        # Parse split rows to intended format
        structured_doc = {}
        structured_doc, remaining_entries = self.root.get_structured_doc_table_rows(
            structure=structure
        )
        if fmt == "markdown":
            for key_path, entries in structured_doc.items():
                structured_doc[key_path] = Doc(entries).markdown_table
            remaining_doc = Doc(remaining_entries).markdown_table
        elif fmt == "yaml":
            for key_path, entries in structured_doc.items():
                structured_doc[key_path] = Doc(entries).nice_yaml
            remaining_doc = Doc(remaining_entries).nice_yaml
        elif fmt == "none":
            for key_path, entries in structured_doc.items():
                structured_doc[key_path] = Doc(entries)
            remaining_doc = Doc(remaining_entries)
        else:
            raise RuntimeError("Unknown documentation format")
        return (structured_doc, remaining_doc)



"""
None data type classes
"""

# pylint: disable=too-few-public-methods

import nadap.types.base

DOC_DT_NAME = "None"
DOC_DT_DESCRIPTION = """
A **none** data type tests data for being python's built-in `None`.
"""
DOC_DT_FEATURES = """
- Supports **Referencing Feature** (even if mostly senseless)
- Data type conversion is not supported
"""
DOC_DT_YAML_EXAMPLE = """
type: none
description: "Example for a none/null definition"
default_value: null
"""


class Null(nadap.types.base.BaseType):
    """
    None type datatype class
    """

    data_type_name = "none"
    _cls_python_classes = [type(None)]
    _convert_to_classes = {}


DOC_DT_CLASS = Null  # pylint: disable=invalid-name



"""
BGP extended community data type class
"""

# pylint: disable=too-few-public-methods

from typing import TYPE_CHECKING
from ipaddress import IPv4Address, AddressValueError

import nadap.types.base
import nadap.mixin.ranges
import nadap.mixin.allowed_value
import nadap.mixin.not_allowed_value
import nadap.schema
from nadap.base import ValEnv, CONVERT_DATA
from nadap.errors import SchemaDefinitionError
from nadap.doc import UnorderedTextList

if TYPE_CHECKING:
    from nadap.doc import TextField


DOC_DT_NAME = "BGP Extended Community"
DOC_DT_DESCRIPTION = """
A **bgp_ext_community** data type tests data for being an instance of
python's built-in class `str` and if it represents a BGP extended standard community (6 bytes).
"""
DOC_DT_FEATURES = """
- Validate BGP Extended Community format
- Supports BGP Extended Community format conversion
- Validate BGP Extended Community specific attributes
- Validate against allowed and not allowed BGP Extended Communities
- Supports **Referencing Feature**
"""
DOC_DT_YAML_EXAMPLE = r"""
type: bgp_ext_community
description: "Example bgp_ext_community definition"
default_value: 10.0.0.1:33

format: ip-address
convert_to_format: asn

not_allowed_values:
  - 10.0.0.0:1

not_allowed_ranges:
  - start: 11.0.0.0:0
    end: 11.255.255.255:65535

reference: ref_key
"""


IP_FORMAT = "ip-address"  # "192.1.1.1:100"
ASN_FORMAT = "asn"  # "65500:33"


class BgpExtCommunity(
    nadap.mixin.not_allowed_value.NotAllowedValueMixin,
    nadap.mixin.allowed_value.AllowedValueMixin,
    nadap.mixin.ranges.ValueRangesMixin,
    nadap.types.base.BaseType,
):
    """
    BGP Community datatype class
    """

    data_type_name = "bgp_ext_community"
    _cls_python_classes = [str]
    _convert_to_classes = {}
    _doc_data_type = "BGP Extended Community"

    def __init__(self, **kwargs):
        self._format = None
        self._detected_format = None
        self._convert_to_format = None
        super().__init__(**kwargs)

    @staticmethod
    def _check_format_str(format_, path):
        if format_ not in [IP_FORMAT, ASN_FORMAT]:
            raise SchemaDefinitionError(  # pylint: disable=raise-missing-from
                msg="Format unknown",
                path=path,
            )

    def _validate_options(self, schema_path: str):
        super()._validate_options(schema_path=schema_path)

        if self._format is not None:
            nadap.schema.is_str(self._format, f"{schema_path}.format")
            self._check_format_str(self._format, f"{schema_path}.format")

        if self._convert_to_format is not None:
            self._check_format_str(
                self._convert_to_format, f"{schema_path}.convert_to_format"
            )

    def _pop_options(self, definition: dict, schema_path: str):
        self._format = definition.pop("format", self._format)
        self._convert_to_format = definition.pop(
            "convert_to_format", self._convert_to_format
        )
        super()._pop_options(definition, schema_path)

    def _apply_data_format(self, data: "any", fmt: str = None) -> "any":
        part1 = int(data / (2**16))
        part2 = data % (2**16)
        if fmt == IP_FORMAT:
            part1 = IPv4Address(part1)
        return f"{part1}:{part2}"

    def _detect_data_format(self, data) -> str:
        p1 = data.split(":")[0]
        if "." in p1:
            return IP_FORMAT
        return ASN_FORMAT

    def _validate_format(
        self,
        data: any,
        path: str,
        env: "ValEnv" = None,
    ) -> str:
        # pylint: disable=too-many-branches
        def raise_fail(path, env):
            self._raise_exceptions(
                "Value is not a valid BGP extended community", path, env
            )

        def raise_asn_fail(path, env):
            self._raise_exceptions(
                "Value is not a valid BGP extended community in ASN format", path, env
            )

        def raise_ip_fail(path, env):
            self._raise_exceptions(
                "Value is not a valid BGP extended community in IP-Address format",
                path,
                env,
            )

        int_value = 0
        if data.count(":") != 1:
            raise_fail(path, env)
        part1, part2 = data.split(":")
        # Analyzing part 1 (before :)
        if part1.count(".") == 3:
            # Test for IP-format
            if self._format == ASN_FORMAT:
                raise_asn_fail(path, env)
            try:
                int_value = int(IPv4Address(part1))
                self._detected_format = IP_FORMAT
            except AddressValueError:
                raise_fail(path, env)
        else:
            # Test for ASN-Format
            if self._format == IP_FORMAT:
                raise_ip_fail(path, env)
            try:
                int_value = int(part1)
                self._detected_format = ASN_FORMAT
            except ValueError:
                raise_fail(path, env)
            if not 0 <= int_value < 2**32:
                raise_fail(path, env)
        int_value = 2**16 * int_value
        try:
            part2_int = int(part2)
        except ValueError:
            raise_fail(path, env)
        if not 0 <= part2_int < 2**16:
            raise_fail(path, env)
        int_value += int(part2)
        return int_value

    def _preprocess_data(self, data: any, env: "ValEnv"):
        if CONVERT_DATA in env.flags and self._convert_to_format is not None:
            data = self._apply_data_format(data, self._convert_to_format)
        else:
            data = self._apply_data_format(data, self._detected_format)
        return data

    def _test_data_type(self, data: str, path: str, env: "ValEnv" = None) -> "str":
        data = super()._test_data_type(data=data, path=path, env=env)
        data = self._validate_format(data=data, path=path, env=env)
        if not 0 <= data < 2**48:
            self._raise_exceptions(
                "BGP extended community value is out of range", path, env
            )
        return data

    @property
    def restrictions(self) -> "TextField":
        """
        Get all restrictions for valid data
        """
        tf = super().restrictions
        _restrictions = UnorderedTextList()
        if self._format:
            _restrictions.append((f"must be in format '{self._format}'"))
        if _restrictions:
            tf.append("BGP Extended Community:")
            tf.append(_restrictions)
        return tf

    @property
    def yaml_data_type(self) -> str:
        """Get YAML data type string"""
        return self._doc_data_type

    @property
    def doc_types(self) -> list[str]:
        """Get list of data type strings"""
        return [self._doc_data_type]

    @classmethod
    def _doc_options_md_upper_part(cls) -> list[str]:
        allowed_formats = (
            "allowed_values:<br>"
            + f"{cls._markdown_indent}- {ASN_FORMAT}<br>"
            + f"{cls._markdown_indent}- {IP_FORMAT}<br>"
        )
        f_desc = (
            f"- {ASN_FORMAT} example: 65500:33<br>"
            + f"- {IP_FORMAT} example: 1.1.1.1:33<br>"
        )
        return super()._doc_options_md_upper_part() + [
            f"| **format** | <code>enum</code> | | {allowed_formats} | {f_desc} |",
            f"| **convert_to_format** | <code>enum</code> | | {allowed_formats} | |",
        ]

    @classmethod
    def _doc_options_yaml_upper_part(cls) -> list[str]:
        return super()._doc_options_yaml_upper_part() + [
            f"format: <{ASN_FORMAT}|{IP_FORMAT}>",
            f"convert_to_format: <{ASN_FORMAT}|{IP_FORMAT}>",
        ]


DOC_DT_CLASS = BgpExtCommunity  # pylint: disable=invalid-name
